import 'package:durt/src/gva/queries.dart';
import 'package:durt/src/gva/transaction.dart';
import 'package:graphql/client.dart';

class Gva {
  final GraphQLClient _client;
  Gva._(this._client);

  /// Main constructor for GVA query object
  factory Gva({required String node}) {
    final _httpLink = HttpLink(node);

    final GraphQLClient client = GraphQLClient(
      cache: GraphQLCache(),
      link: _httpLink,
    );

    return Gva._(client);
  }

  /// Return the balance from a given pubkey, or 0 if null.
  Future<double> balance(String pubkey, {bool ud = false}) async {
    double balance = 0;

    final QueryOptions options = QueryOptions(
      document: gql(getBalanceQuery),
      variables: <String, dynamic>{
        'pubkey': pubkey,
      },
    );
    final QueryResult result = await _client.query(options);

    if (result.hasException) {
      print(result.exception.toString());
      throw Exception('Unexpected error happened in graphql request');
    } else if (result.data?['balance'] != null) {
      balance = result.data!['balance']['amount'] / 100;
      if (ud) {
        double currentUd = result.data!['currentUd']['amount'] / 100;
        balance = balance / currentUd;
      }
    }

    return balance.toPrecision(2);
  }

  /// Return the username from a given pubkey.
  Future<String> getUsername(String pubkey) async {
    String username = '';

    final QueryOptions options = QueryOptions(
      document: gql(getIdQuery),
      variables: <String, dynamic>{
        'pubkey': pubkey,
      },
    );
    final QueryResult result = await _client.query(options);

    if (result.hasException) {
      print(result.exception.toString());
      throw Exception('Unexpected error happened in graphql request');
    } else if (result.data?['idty']?['username'] != null) {
      username = result.data!['idty']['username'];
    }

    return username;
  }

  /// Return the username from a given pubkey.
  Future<double> getCurrentUd() async {
    final QueryOptions options = QueryOptions(document: gql(currentUdQuery));
    final QueryResult result = await _client.query(options);

    if (result.hasException) {
      print(result.exception.toString());
      throw Exception('Unexpected error happened in graphql request');
    } else {
      return int.parse(result.data!['currentUd']['amount']) / 100;
    }
  }

  /// Pay a given pubkey.
  /// This automatically recognizes whether it is an HD wallet or a Cesium wallet.
  /// Return 'success' string if the transaction has been successfuly proced, or the specified error
  Future<String> pay(
      {required String recipient,
      required double amount,
      String? dewif,
      String? password,
      String? mnemonic,
      String? cesiumId,
      String? cesiumPwd,
      String lang = 'english',
      int? derivation,
      String comment = '',
      bool ud = false,
      bool useMempool = false}) async {
    Transaction transaction;

    if (ud) {
      final currentUd = await getCurrentUd();
      amount = amount * currentUd;
    }

    transaction = Transaction(
        recipient: recipient,
        amount: (amount.toPrecision(2) * 100).toInt(),
        dewif: dewif,
        password: password,
        mnemonic: mnemonic,
        cesiumId: cesiumId,
        cesiumPwd: cesiumPwd,
        derivation: derivation ?? -1,
        comment: comment,
        useMempool: useMempool,
        lang: lang,
        client: _client);

    // Execute transaction
    return await transaction.process();
  }
}

extension Ex on double {
  double toPrecision(int n) => double.parse(toStringAsFixed(n));
}
